local AddonName, GGL = ...
local Templates = GGL.Templates

local TEMPLATE_NAME = "BNPlayer"

local BNPlayer, BNPlayerHelper = Templates:Register(TEMPLATE_NAME)

-- https://wow.gamepedia.com/World_of_Warcraft_API#Battle.net

-- lua
local assert = assert
local match = string.match
local mathmin = math.min

local C_BattleNet = C_BattleNet or GGL.C_BattleNet

-- wow
local BNGetFriendAccountInfo, BNGetAccountInfoByID = C_BattleNet.GetFriendAccountInfo, C_BattleNet.GetAccountInfoByID
local BNGetFriendNumGameAccounts, BNGetFriendGameAccountInfo = C_BattleNet.GetFriendNumGameAccounts, C_BattleNet.GetFriendGameAccountInfo
local BNGetNumFriends, BNGetFriendIndex, BNGetInfo = BNGetNumFriends, BNGetFriendIndex, BNGetInfo
local BNGetGameAccountInfoByID = C_BattleNet.GetGameAccountInfoByID

local BNET_CLIENT_WOW = _G.BNET_CLIENT_WOW -- "WoW"
local WOW_PROJECT_ID = _G.WOW_PROJECT_ID or 1
local WOW_PROJECT_MAINLINE = _G.WOW_PROJECT_MAINLINE or 1
local WOW_PROJECT_CLASSIC = _G.WOW_PROJECT_CLASSIC or 2
local WOW_PROJECT_BURNING_CRUSADE_CLASSIC = _G.WOW_PROJECT_BURNING_CRUSADE_CLASSIC or 5
local WOW_PROJECT_WRATH_CLASSIC = _G.WOW_PROJECT_WRATH_CLASSIC or 11

local ALLOWED_PROJECT_ID = {
    [WOW_PROJECT_MAINLINE]                      = true,
    [WOW_PROJECT_WRATH_CLASSIC]                 = true,
    [WOW_PROJECT_BURNING_CRUSADE_CLASSIC]       = true,
    [WOW_PROJECT_CLASSIC]                       = true,
}

local GetTime = GGL.GetTime

-- ##################
-- local
-- ##################
local ADDON_PREFIX = "GGL_ADDON_MSG"
local AccountRegister = {}
local AccountRegister_BattleTag = {}
local WoWGuid_to_Bnet = {}

-- ##################
-- static
--
-- EVENTS:
--      "PlayerTemplate_PlayerUpdate"              (bnPlayerTemplate)
--      "PlayerAcountIDUpdate"      (Templates.BNPlayer, updatetSyncIDs)
-- ##################
local Static = {}
Templates.BNPlayer = Static

function Static:Get(bnetAccountID)
    return ( bnetAccountID and AccountRegister[bnetAccountID] ) and AccountRegister[bnetAccountID] or nil
end

function Static:UpdateByID(bnetAccountID)
    if not bnetAccountID then return end
    if not AccountRegister[bnetAccountID] then
        -- Add new one
        Templates:New(TEMPLATE_NAME, bnetAccountID)
    else
        AccountRegister[bnetAccountID]:Update()
    end
end

function Static:UpdateByIndex(bnetAccountIndex)
    if not bnetAccountIndex then return end
    local accountInfo = BNGetFriendAccountInfo(bnetAccountIndex)
    if accountInfo then
        Static:UpdateByID(accountInfo.bnetAccountID)
    end
end

-- Update all accounts
function Static:UpdateAll()
    local numAccounts = BNGetNumFriends()
    if numAccounts and numAccounts > 0 then
        for i = 1, numAccounts do
            Static:UpdateByIndex(i)
        end
    end
end

function Static:GetWoWGUIDList()
    return WoWGuid_to_Bnet
end

function Static:GetByWoWGUID(wowPlayerGUID)
    if not wowPlayerGUID or not WoWGuid_to_Bnet[wowPlayerGUID] then return GGL:AddIntoErrorLog("BNPlayer / Static:GetByWoWGUID", wowPlayerGUID or "nil") end
    return WoWGuid_to_Bnet[wowPlayerGUID]
end

function Static:GetSelf()
    local _, battleTag, toonID = BNGetInfo()
    if not AccountRegister_BattleTag[battleTag] and toonID then
        Static:UpdateByID(toonID)
    end
    return AccountRegister_BattleTag[battleTag]
end

function Static:GetLastSeenForBattleTag(battleTag)
    if not battleTag then return end
    return GGL.db.BTagLastSeen[battleTag]
end

--- method(BNPlayer)
function Static:RegisterCallbackEvent(...)
    BNPlayerHelper:RegisterCallbackEvent(...)
end

function Static:UnregisterCallbackEvent(...)
    BNPlayerHelper:UnregisterCallbackEvent(...)
end

-- ##################
-- Watcher
-- ##################
function BNPlayerHelper:BN_FRIEND_INFO_CHANGED(bnetAccountIndex)
    if bnetAccountIndex then
        Static:UpdateByIndex(bnetAccountIndex)
    end
end

-- fire on add/remove friend. AccountID dont get removed.
function BNPlayerHelper:BN_FRIEND_LIST_SIZE_CHANGED(bnetAccountID)
    if bnetAccountID then
        Static:UpdateByID(bnetAccountID)
    end
end

function BNPlayerHelper:FRIENDLIST_UPDATE()
    Static:UpdateAll()
end

-- ##################
-- Init
-- ##################
local RegisterAccountID
do
    -- BattleTag#123:UnitGUID
    local SYNC_ID_FORMAT = "^(.*):(.*)"

    -- [battleTag] = {
    --          [accountID] = { ["GUID"] = dbTable },
    -- }
    local BnetAccountIDRegister = {}

    -- [gameAccountID] = { accountID = <accountID>, [playerTemplate] = true }
    local BnetGameAccountIDs = {}

    local function GetNextAccountID(t)
        for i = 1, 9999 do
            if not t[i] then
                return i
            end
        end
    end

    local function InitRegister()

        -- BnetAccountIDRegister
        for syncID, playerData in pairs(GGL.db.PlayerRegister) do
            local battleTag, unitGUID = match(syncID, SYNC_ID_FORMAT)
            if battleTag then
                if not BnetAccountIDRegister[battleTag] then BnetAccountIDRegister[battleTag] = {} end

                -- only entrys with accountID are tracked
                if playerData.accountID then
                    -- add accountID table
                    if not BnetAccountIDRegister[battleTag][playerData.accountID] then BnetAccountIDRegister[battleTag][playerData.accountID] = {} end
                    -- add guid
                    BnetAccountIDRegister[battleTag][playerData.accountID][unitGUID] = playerData
                end
            end
        end

        -- only needed once
        InitRegister = nil
    end

    local function UpdateAccountID(gameAccountID, battleTag, oldAccountID, newAccountID)
        -- create a update list with playerSyncIDs that get an update
        local updateList = {}
        -- first update all templates with the new account ID
        if not BnetAccountIDRegister[battleTag][newAccountID] then BnetAccountIDRegister[battleTag][newAccountID] = {} end
        for playerTemplate, isPlayerTemplate in pairs(BnetGameAccountIDs[gameAccountID]) do
            -- check if entry is a playerTemplate
            if isPlayerTemplate == true then
                if playerTemplate:SetAccountID(newAccountID) then
                    updateList[playerTemplate.syncID] = newAccountID
                end
                BnetAccountIDRegister[battleTag][newAccountID][playerTemplate.unitGUID] = playerTemplate:GetDBTable()
            end
        end

        -- now update all not loaded players
        if oldAccountID and BnetAccountIDRegister[battleTag][oldAccountID] then
            -- check for existing target accountID or create new table
            local newAccountIDRegister = BnetAccountIDRegister[battleTag][newAccountID] or {}
            for unitGUID, playerData in pairs(BnetAccountIDRegister[battleTag][oldAccountID]) do
                -- check if its a new accountID and set it
                if not playerData.accountID or playerData.accountID ~= newAccountID then
                    updateList[battleTag..":"..unitGUID] = newAccountID
                    playerData.accountID = newAccountID
                end
                newAccountIDRegister[unitGUID] = playerData
            end
            -- set new accountID table and delete old one
            BnetAccountIDRegister[battleTag][newAccountID] = newAccountIDRegister
            BnetAccountIDRegister[battleTag][oldAccountID] = nil
        end

        -- set accountID for gameAccountID
        BnetGameAccountIDs[gameAccountID].accountID = newAccountID

        BNPlayerHelper:FireCallbackEvent("PlayerAcountIDUpdate", Static, updateList)
    end

    function RegisterAccountID(playerTemplate)
        if InitRegister then InitRegister() end
        if not playerTemplate then return end

        local battleTag = playerTemplate.bnPlayer.battleTag
        -- new battle tag??
        if not BnetAccountIDRegister[battleTag] then BnetAccountIDRegister[battleTag] = {} end
        local bTagRegister = BnetAccountIDRegister[battleTag]
        -- get accountIDs
        local gameAccountID = playerTemplate.gameAccountID  -- blizzard
        local accountID = playerTemplate:GetAccountID()     -- GGL

        if not BnetGameAccountIDs[gameAccountID] then
            -- new entry nothing to check here
            BnetGameAccountIDs[gameAccountID] = {
                [playerTemplate] = true,
                accountID = accountID,
            }
        elseif not BnetGameAccountIDs[gameAccountID][playerTemplate] then
            BnetGameAccountIDs[gameAccountID][playerTemplate] = true

            -- Update AccountID
            if accountID and not BnetGameAccountIDs[gameAccountID].accountID then
                UpdateAccountID(gameAccountID, battleTag, nil, accountID)
            elseif accountID and BnetGameAccountIDs[gameAccountID].accountID ~= accountID then
                -- two different accountID's found so we get the lowest and set it for all
                UpdateAccountID(gameAccountID, battleTag, BnetGameAccountIDs[gameAccountID].accountID, mathmin(BnetGameAccountIDs[gameAccountID].accountID, accountID))
            elseif not accountID and not BnetGameAccountIDs[gameAccountID].accountID then
                -- no accountID found but at least 2 chars so we set one
                UpdateAccountID(gameAccountID, battleTag, nil, GetNextAccountID(bTagRegister))
            elseif not accountID and BnetGameAccountIDs[gameAccountID].accountID then
                -- no accountID fount but there is a existing one
                UpdateAccountID(gameAccountID, battleTag, nil, BnetGameAccountIDs[gameAccountID].accountID)
            else
                UpdateAccountID(gameAccountID, battleTag, nil, accountID or GetNextAccountID(bTagRegister))
            end
        end
    end

end

local function OnInit()
    RegisterAccountID()

    Static:UpdateAll()

    BNPlayerHelper:RegisterEvent("BN_FRIEND_LIST_SIZE_CHANGED")
    BNPlayerHelper:RegisterEvent("BN_FRIEND_INFO_CHANGED")
    if not GGL.IS_RETAIL then
        BNPlayerHelper:RegisterEvent("FRIENDLIST_UPDATE")
    end

    Static:GetSelf()
end
BNPlayerHelper:AddOnInit(OnInit)

-- ##################
-- template
-- Variable
--      bnetAccountID
--      battleTag
-- Functions
-- ##################
function BNPlayer:Constructor(bnetAccountID)
    assert(bnetAccountID, "No 'bnetAccountID' found ("..(bnetAccountID or "nil")..")")

    if AccountRegister[bnetAccountID] then
        return AccountRegister[bnetAccountID]
    end

    -- https://wow.gamepedia.com/API_C_BattleNet.GetAccountInfoByID
    local accountInfo = BNGetAccountInfoByID(bnetAccountID)

    if not accountInfo then return end

    -- some static things
    self.bnetAccountID = bnetAccountID
    self.battleTag = accountInfo.battleTag

    -- Save for usage
    AccountRegister[bnetAccountID] = self
    AccountRegister_BattleTag[self.battleTag] = self

    self:Update()
end

function BNPlayer:SetLastSeen(lastSeenTime)
    if not self.battleTag then return end
    lastSeenTime = lastSeenTime or GetTime()
    GGL.db.BTagLastSeen[self.battleTag] = lastSeenTime
end

function BNPlayer:GetLastSeen()
    if not self.battleTag then return end
    return GGL.db.BTagLastSeen[self.battleTag]
end

local function AddBNWoWPlayer(self, unitGUID)
    if not unitGUID then return end
    if not WoWGuid_to_Bnet[unitGUID] then
        if not self.wowPlayerList then self.wowPlayerList = {} end
        if not self.wowPlayerAccountList then self.wowPlayerAccountList = {} end
        local wowPlayer = Templates:Get("BNWoWPlayer", unitGUID, self)
        self.wowPlayerList[#self.wowPlayerList+1] = wowPlayer
        WoWGuid_to_Bnet[unitGUID] = wowPlayer
        -- check if we have this account alredy
        RegisterAccountID(wowPlayer)
    else
        WoWGuid_to_Bnet[unitGUID]:Update()
    end
end

function BNPlayer:Update()
    local accountInfo = BNGetAccountInfoByID(self.bnetAccountID)
    local accountIndex = BNGetFriendIndex(self.bnetAccountID)
    local _, ownBattleTag, ownToonID = BNGetInfo()

    -- set data
    self.isFriend = accountInfo.isFriend
    --self.isOffline = accountInfo.lastOnlineTime and true or false
    self.isAFK = accountInfo.isAFK

    -- update old wow accounts
    local old = {}
    if self.wowPlayerList and #self.wowPlayerList > 0 then
        for i = 1, #self.wowPlayerList do
            self.wowPlayerList[i]:Update()
            old[self.wowPlayerList[i].unitGUID] = true
        end
    end

    -- check for new wow chars
    if ownBattleTag == self.battleTag then
        -- Player
        local playerGuid = UnitGUID("player")
        if playerGuid then
            AddBNWoWPlayer(self, playerGuid)
        end
    elseif accountIndex and accountIndex > 0 then
        local numGameAccounts = BNGetFriendNumGameAccounts(accountIndex)
        if numGameAccounts and numGameAccounts > 0 then
            for i = 1, numGameAccounts do
                local gameAccount = BNGetFriendGameAccountInfo(accountIndex, i)
                if gameAccount and gameAccount.clientProgram == BNET_CLIENT_WOW and gameAccount.wowProjectID and ALLOWED_PROJECT_ID[gameAccount.wowProjectID] and gameAccount.playerGuid and not old[gameAccount.playerGuid] then
                    AddBNWoWPlayer(self, gameAccount.playerGuid)
                end
            end
        end
    elseif not accountIndex and accountInfo and accountInfo.gameAccountInfo and accountInfo.gameAccountInfo.gameAccountID then
        local gameAccount = BNGetGameAccountInfoByID(accountInfo.gameAccountInfo.gameAccountID)
        if gameAccount and gameAccount.clientProgram == BNET_CLIENT_WOW and gameAccount.wowProjectID and ALLOWED_PROJECT_ID[gameAccount.wowProjectID] and gameAccount.playerGuid and not old[gameAccount.playerGuid] then
            AddBNWoWPlayer(self, gameAccount.playerGuid)
        end
    end

    BNPlayerHelper:FireCallbackEvent("PlayerTemplate_PlayerUpdate", self)
end
