local AddonName, GGL = ...
local Templates = GGL.Templates

local Session, SessionHelper = Templates:Register("Session")

-- lua
local tonumber = tonumber
local pairs, ipairs = pairs, ipairs
local format, match, str_sub, str_split, str_match = string.format, string.match, string.sub, string.split, string.match
local time = time

-- wow
local C_TimerAfter = C_Timer.After

-- ##################
-- local
-- ##################
local RunningSessions = {}

local SESSION_FORMAT = "S%d"
local TIMER_INTERVAL = 1

-- lootTime, zoneIniID, difficultyID, itemID, ammount, sourceID, mainSourceFlag, subSourceFlag, mapX, mapY
local ITEM_DB_FORMAT = "%d\001%d\001%d\001%s\001%d\001%d\001%d\001%d\001%d\001%d"
local ITEM_DB_FORMAT_MATCH = "(%d*)\001(%d*)\001(%d*)\001(.*)\001(%d*)\001(%d*)\001(%d*)\001(%d*)\001(%d*)\001(%d*)"
local PLAYER_FORMAT = "%s:%s"

local function AddLogString(self, playerTemplate, event, source, db, lootTime, zoneID, difficultyID, itemLog, mapX, mapY)
    if lootTime < 0 then return end
    itemLog = tostring(itemLog)

    if source then
        -- source(mainSourceFlag, subSourceFlag, categoryFlag, sourceID, quant)
        for i, sourceEntry in ipairs(source) do
            if sourceEntry and type(sourceEntry) == "table" then
                db[ #db + 1 ] = format(ITEM_DB_FORMAT, lootTime, zoneID, difficultyID, itemLog, sourceEntry[5], sourceEntry[4] or 0, sourceEntry[1], sourceEntry[2], mapX or 0, mapY or 0)
                self:FireCallbackEvent(event, self, playerTemplate, db[ #db ], #db)
            end
        end
    end
end

-- ##################
-- static
-- ##################
local Static = {}
Templates.Session = Static

-- combine by mainSourceFlag
function Static:FormatSource(source)
    local tmp = {}
    -- source(mainSourceFlag, sourceID, ammount)
    for index, src in ipairs(source) do
        tmp[src[1]] = ( tmp[src[1]] or 0 ) + src[3]
    end
    return tmp
end

function Static:GetRunningSession(sessionDB)
    sessionDB = sessionDB or GGL.sessionDb.Sessions
    if sessionDB.RunningSession then
        return Templates:New("Session", sessionDB, sessionDB.RunningSession)
    end
end

function Static:CleanUpDB()
    local curTime = GGL.GetTime()
    local sessionDB = GGL.sessionDb.Sessions
    local cleanSessions = {}
    for sessionID, sessionData in pairs(sessionDB) do
        if not sessionData.wasStarted then
           cleanSessions[#cleanSessions+1] = sessionID
        elseif sessionData.endTime and curTime > sessionData.endTime then
            -- archive the old session
            GGL:AddSessionIntoArchive(sessionData)
            -- remove from db
           cleanSessions[#cleanSessions+1] = sessionID
            GGL:AddIntoErrorLog("Session:CleanUpDB / archive old session", sessionID)
        elseif GGL.SessionRegister then
            local sessionType = str_sub(sessionID, 1, 1)
            -- Solo session cleanup
            if sessionType == "A" and GGL.SessionRegister[sessionType] ~= sessionID then
                -- looks like something went wrong with that season so we delete it
               cleanSessions[#cleanSessions+1] = sessionID
                GGL:AddIntoErrorLog("Session:CleanUpDB / unknown session", sessionID)
            elseif GGL.SessionRegister[sessionType] ~= sessionID and sessionData.endlessTime and curTime > (sessionData.startTime + (60 * 60 * 48)) then
                -- A endless Session that was startet over 2 days ago should be also moved into the archive
                -- archive the old session
                GGL:AddSessionIntoArchive(sessionData)
                -- remove from db
               cleanSessions[#cleanSessions+1] = sessionID
                GGL:AddIntoErrorLog("Session:CleanUpDB / archive old session", sessionID)
            elseif GGL.SessionRegister[sessionType] ~= sessionID and not sessionData.startTime then
               cleanSessions[#cleanSessions+1] = sessionID
                GGL:AddIntoErrorLog("Session:CleanUpDB / unknown session", sessionID)
            end
        end
    end
    if #cleanSessions > 0 then
        for i = 1, #cleanSessions do
            sessionDB[cleanSessions[i]] = nil
        end
    end
end

function Static:CloseOpenSessions()
    for k,v in pairs(RunningSessions) do
        v:Close()
    end
end

function Static:GetDataFromLogString(logString)
    local lootTime, zoneIniID, difficultyID, itemLog, ammount, sourceID, mainSourceFlag, subSourceFlag, mapX, mapY = match(logString, ITEM_DB_FORMAT_MATCH)
    local itemID = GGL:GetItemDataFromItemLog(itemLog)
    return tonumber(lootTime), tonumber(zoneIniID), tonumber(difficultyID), tonumber(itemID), itemLog, tonumber(ammount), tonumber(sourceID), tonumber(mainSourceFlag), tonumber(subSourceFlag), tonumber(mapX), tonumber(mapY)
end

-- ##################
-- EVENTS
--      "SessionTemplate_AddItem"           (sessionSelf, itemID, itemString, zoneID, sourceID, ammount, copperValue, mainSourceFlag, subSourceFlag, sourcePlayer)
--      "SessionTemplate_AddMoney"          (sessionSelf, copper, zoneID, sourceID, mainSourceFlag, subSourceFlag, sourcePlayer)
--      "SessionTemplate_AddLogItem"        (sessionSelf, playerTemplate, logString, logCounter)
--      "SessionTemplate_AddLogMoney"       (sessionSelf, playerTemplate, logString, logCounter)
--      "SessionTemplate_SessionClosed"     (sessionSelf)
--      "SessionTemplate_SessionStopped"    (sessionSelf)
--      "SessionTemplate_SetDataRefresh"    (sessionSelf)
-- ##################

Session.sessionFormat = "S%d"
Session.sessionType = "S"

--- New session or reload one
function Session:Constructor(sessionDB, sessionID)
    if not sessionDB or type(sessionDB) ~= "table" then return end

    -- check for running session or create new one
    if not sessionID then
        sessionID = format(self.sessionFormat, GGL.GetTime())
    end
    if RunningSessions[sessionID] then
        return RunningSessions[sessionID]
    end
    if not sessionDB[sessionID] then
        sessionDB[sessionID] = {}
    end

    -- set template variables
    self.dbRaw = sessionDB
    self.db = sessionDB[sessionID]
    -- set db variables
    self.db.sessionID = sessionID   -- double the fun
    self.dbRaw.RunningSession = sessionID
    -- add own callback handler
    GGL:AddCallbackHandler(self)

    local db = self.db

    -- if its a running session resume or end it
    if (db.isRunning or db.isStopped) or (db.endTime and db.endTime <= GGL.GetTime()) then
        if db.endTime and db.endTime <= GGL.GetTime() then
            return self:Close()
        else
            self:SetData()
        end
    elseif db.isClosed then
        return self:Close()
    else
        -- new one
    end

    -- list for syncID's
    self.mergeIDList = {}

    -- create lootLog table
    db.lootLog = db.lootLog or {}

    -- set own bnet
    self.selfPlayer = GGL.Templates.BNWoWPlayer:GetSelfPlayer()
    self.selfDbKey = self:GetPlayerSyncID(self.selfPlayer)
    if not db.lootLog[self.selfDbKey] then db.lootLog[self.selfDbKey] = {} end
    self.lootLog = db.lootLog[self.selfDbKey]

    -- Register loot module
    self.OnItemLoot_CB = function(Loot, itemID, itemLog, ammount, source)
        if not db.isRunning then return end
        self:AddNewItem(itemID, itemLog, ammount, source)
    end

    self.OnMoneyLoot_CB = function(Loot, ammount, source)
        if not db.isRunning then return end
        self:AddMoney(ammount, source)
    end

    -- Register zone change
    self.OnMapChange_CB = function(sender, mapID, mapName, mapTypeFlag, difficultyID)
        self.zoneID = mapID
        self.zoneDifficultyID = difficultyID
    end
    self.OnMapChange_CB(nil, GGL:GetZoneData())

    -- Register kill log
    db.playerKillCount = db.playerKillCount or {}
    db.ownKillCount = db.ownKillCount or 0
    self.OnPlayerKillNpc_CB = function(npcID, npcName)
        if not db.isRunning then return end
        if not db.playerKillCount[self.selfDbKey] then db.playerKillCount[self.selfDbKey] = 0 end
        db.playerKillCount[self.selfDbKey] = db.playerKillCount[self.selfDbKey] + 1

        if self.mainFrame then
            -- just add one kill
            db.ownKillCount = db.ownKillCount + 1
            self.mainFrame:SetKillCount(db.ownKillCount)
        end
    end

    -- Set Price table
    if not db.priceLog then db.priceLog = {} end

    -- Add into statistics
    self.statistics = GGL.Statistics:RegisterSession(self)

    -- db instanceResetData
    self.db.instanceResetData = self.db.instanceResetData or {}

    -- register as running
    RunningSessions[sessionID] = self
end

function Session:GetPlayerSyncID()
    -- the player gets a fixed mergeID
    -- TODO: #9 Add more merge settings here
    if not self.selfPlayer then
        self.selfPlayer = GGL.Templates.BNWoWPlayer:GetSelfPlayer()
    end
    return self.selfPlayer.syncID
end

--- Returns the merge ID for a playerTemplate
--- For session that returns just "self"
function Session:GetPlayerMergeID(playerTemplate)
    return "self"
end

--- Returns the active player for a mergeID
--- For session that returns just the player
-- @return  playerTemplate
function Session:GetActivePlayerForMergeID(mergeID)
    return self.selfPlayer
end

--- Returns the merge list
-- format: mergeList = { [playerSyncID] = <0=noID or "string"=mergeID> }
function Session:GetMergeList()
    return self.mergeIDList
end

-- lootTime, zoneIniID, difficultyID, itemID, itemLog, ammount, sourceID, mainSourceFlag, subSourceFlag, mapX, mapY
function Session:GetDataFromLogString(logString)
    local lootTime, zoneIniID, difficultyID, itemLog, ammount, sourceID, mainSourceFlag, subSourceFlag, mapX, mapY = match(logString, ITEM_DB_FORMAT_MATCH)
    local itemID = GGL:GetItemDataFromItemLog(itemLog)
    return tonumber(lootTime), tonumber(zoneIniID), tonumber(difficultyID), tonumber(itemID), itemLog, tonumber(ammount), tonumber(sourceID), tonumber(mainSourceFlag), tonumber(subSourceFlag), tonumber(mapX), tonumber(mapY)
end

-- sessionSelf, itemID, zoneID, sourceID, ammount, copperValue, mainSourceFlag, subSourceFlag, sourcePlayer
local function ProcessMoneyData(self, copper, sourceID, zoneID, difficultyID, mainSourceFlag, subSourceFlag, categoryFlag, sourcePlayer)
    if (self.ItemPrice and self.ItemPrice:RawGoldAllowed())
        and (not self.ZoneFilter or self.ZoneFilter:IsValidZone(zoneID, difficultyID, mainSourceFlag, subSourceFlag))
        and (not self.LootSourceFilter or self.LootSourceFilter:IsValidSource(mainSourceFlag, subSourceFlag)) then

        self:FireCallbackEvent("SessionTemplate_AddMoney", self, copper, zoneID, sourceID, mainSourceFlag, subSourceFlag, sourcePlayer)
    end
end

-- direct call
function Session:ProcessMoneyLootData(player, lootTime, zoneID, difficultyID, copper, sourceID, mainSourceFlag, subSourceFlag, mapX, mapY)
    ProcessMoneyData(self, copper, sourceID, zoneID, difficultyID, mainSourceFlag, subSourceFlag, nil, player)
end

function Session:AddMoney(ammount, source)
    if not self.dataIsSet then return end
    -- lootLog(dbTable) and lootTime check
    local lootTime = time()
    lootTime = lootTime - self.db.startTime
    -- get loot zone
    local zoneID = self.zoneID
    local difficultyID = self.zoneDifficultyID
    local mapX, mapY = GGL:GetCurrentMapXY()

    -- Add loot string
    AddLogString(self, self.selfPlayer, "SessionTemplate_AddLogMoney", source, self.lootLog, lootTime, zoneID, difficultyID, 0, mapX, mapY)

    if source then
        -- source(mainSourceFlag, subSourceFlag, categoryFlag, sourceID, quant)
        for i, sourceEntry in ipairs(source) do
            if sourceEntry and type(sourceEntry) == "table" then
                ProcessMoneyData(self, sourceEntry[5], sourceEntry[4], zoneID, difficultyID, sourceEntry[1], sourceEntry[2], sourceEntry[3], self.selfPlayer)
            end
        end
    end
end

-- sessionSelf, itemID, zoneID, sourceID, ammount, copperValue, mainSourceFlag, subSourceFlag, sourcePlayer
local function ProcessItemData(self, itemID, itemString, ammount, price, sourceID, zoneID, difficultyID, mainSourceFlag, subSourceFlag, categoryFlag, sourcePlayer)
    if (self.ItemFilter and self.ItemFilter:IsValidItem(itemID, itemString))
        and (not self.ZoneFilter or self.ZoneFilter:IsValidZone(zoneID, difficultyID, mainSourceFlag, subSourceFlag))
        and (not self.LootSourceFilter or self.LootSourceFilter:IsValidSource(mainSourceFlag, subSourceFlag)) then

        self:FireCallbackEvent("SessionTemplate_AddItem", self, itemID, itemString, zoneID, sourceID, ammount, price, mainSourceFlag, subSourceFlag, sourcePlayer)
        GGL.LootAlert:ProcessAlerts(itemID, self:LoggedItemUseVariations(itemID, itemString) and itemString or nil, price, sourcePlayer)
    end
end

-- direct call
function Session:ProcessItemLootData(player, lootTime, zoneID, difficultyID, itemID, itemLog, ammount, sourceID, mainSourceFlag, subSourceFlag, mapX, mapY)
    local newItemID, itemString = GGL:GetItemDataFromItemLog(itemLog)
    local price = self.ItemPrice:GetPrice(newItemID, itemString)
    price = price or 0
    ProcessItemData(self, itemID, itemString, ammount, price, sourceID, zoneID, difficultyID, mainSourceFlag, subSourceFlag, nil, player)
end

function Session:AddNewItem(itemID, itemLog, ammount, source)
    if not self.dataIsSet then return end
    -- lootLog and lootTime check
    local lootTime = time()
    lootTime = lootTime - self.db.startTime
    -- save price for the log
    local newItemID, itemString = GGL:GetItemDataFromItemLog(itemLog)
    local price = self.ItemPrice:GetPrice(newItemID, itemString)
    price = price or 0

    -- get loot zone
    local zoneID = self.zoneID
    local difficultyID = self.zoneDifficultyID
    local mapX, mapY = GGL:GetCurrentMapXY()

    -- always into log
    AddLogString(self, self.selfPlayer, "SessionTemplate_AddLogItem", source, self.lootLog, lootTime, zoneID, difficultyID, itemLog,  mapX, mapY)

    if source then
        -- source(mainSourceFlag, subSourceFlag, categoryFlag, sourceID, quant)
        for i, sourceEntry in ipairs(source) do
            if sourceEntry and type(sourceEntry) == "table" then
                ProcessItemData(self, itemID, itemString, sourceEntry[5], price, sourceEntry[4], zoneID, difficultyID, sourceEntry[1], sourceEntry[2], sourceEntry[3], self.selfPlayer)
            end
        end
    end
end

function Session:AddFromLogString(logString)
    self.pauseLog = true
    local lootTime, zoneIniID, difficultyID, itemID, itemLog, ammount, sourceID, mainSourceFlag, subSourceFlag = self:GetDataFromLogString(logString)
    if itemID == 0 then
        ProcessMoneyData(self, ammount, sourceID, zoneIniID, difficultyID, mainSourceFlag, subSourceFlag, nil, self.selfPlayer)
    else
        local newItemID, itemString = GGL:GetItemDataFromItemLog(itemLog)
        local price = self.ItemPrice:GetPrice(newItemID, itemString)
        price = price or 0

        GGL:IsItemCached(itemID, ProcessItemData, self, itemID, itemString, ammount, price, sourceID, zoneIniID, difficultyID, mainSourceFlag, subSourceFlag, nil, self.selfPlayer)
    end
    self.pauseLog = nil
end

function Session:RefreshSettingsSet(setRefresh, priceRefresh)
    local db = self.db
    if db.setSave then
        if setRefresh and self.ItemPrice then
            self.ItemPrice:LoadSetSettings(db.setSave.itemPrice, self.db.priceLog)
            self.ItemFilter:LoadSetSettings(db.setSave.ItemFilter)
            self.ZoneFilter:LoadSetSettings(db.setSave.zoneFilter, db.setSave.instanceFilter)
            self.LootSourceFilter:LoadSetSettings(db.setSave.lootSourceFilter)
        else
            self.ItemPrice = Templates:New("ItemPriceSet", db.setSave.itemPrice, true, self.db.priceLog)
            self.ItemFilter = Templates:New("ItemFilter", db.setSave.ItemFilter, true)
            self.ZoneFilter = Templates:New("ZoneFilter", db.setSave.zoneFilter, db.setSave.instanceFilter, true)
            self.LootSourceFilter = Templates:New("LootSourceFilter", db.setSave.lootSourceFilter, true)
        end
    end
    -- refresh cached prices
    -- wipe(db.priceLog)
    if priceRefresh then
        for itemID in pairs(db.priceLog) do
            self.ItemPrice:GetPrice(GGL:GetItemDataFromItemLog(itemID))
        end
    end
    -- check if we have the mainframe
    if self.mainFrame then
        self:AddIntoMainFrame(self.mainFrame)
    end
end

function Session:SetData(set, setRefresh)
    -- change data of running session not allowed
    local db = self.db

    -- Setdata
    set = set or "Base"
    if type(set) == "table" then
        db.setSave = set
        db.set = false
    else
        if set == true or db.set == false then

        elseif not db.set or setRefresh then
            local SetData = Templates:New("Set", set)
            db.setSave = {
                itemPrice = SetData:GetPriceSource(true),
                ItemFilter = SetData:GetItemFilter(true),
                zoneFilter = SetData:GetZoneFilter(true),
                instanceFilter = SetData:GetInstanceFilter(true),
                lootSourceFilter = SetData:GetLootSourceFilter(true),
            }
        end
        db.set = db.set or set
    end

    -- set player name
    self.playerName = format("%s-%s", UnitName("player"),  GetNormalizedRealmName())

    -- Register/Refresh set
    self:RefreshSettingsSet(setRefresh)

    self.db.dataIsSet = true
    self.dataIsSet = true
end

function Session:ReloadSet()
    if self.db.setSave then
        self:RefreshSettingsSet(true, true)

        -- Fire event before the mainFrame refresh
        -- so other modules get the AddFromLogString from mainFrame also
        self:FireCallbackEvent("SessionTemplate_SetDataRefresh", self)

        if self.mainFrame then
            self:RemoveFromMainFrame()
            self:AddIntoMainFrame(true)
            self:ReloadOwnItemsFromLog()
        end
    end
end

function Session:MainFrameOnNewItem(itemID, itemString, zoneID, sourceID, ammount, copperValue, mainSourceFlag, subSourceFlag, sourcePlayer)
    if not self.mainFrame or sourcePlayer ~= self.selfPlayer then return end
    self.mainFrame:AddItem(itemID, self:LoggedItemUseVariations(itemID, itemString) and itemString or nil, sourceID, ammount, copperValue * ammount, mainSourceFlag, subSourceFlag)
    self.mainFrame:SetItemValueGold(copperValue * ammount, true)
end

function Session:MainFrameOnAddMoney(copper, zoneID, sourceID, mainSourceFlag, subSourceFlag, sourcePlayer)
    if not self.mainFrame or sourcePlayer ~= self.selfPlayer then return end
    self.mainFrame:SetRawGold(copper, true)
end

function Session:AddIntoMainFrame(add)
    if not self.db or self.db.isClosed then return end
    if add then
        local frame = GGL.GUI.MainWindow.frame or GGL.GUI.MainWindow:CreateFrame()
        frame:Clear()
        local db = self.db
        -- start timer if its a running session or just set timer
        if self.db.endlessTime  then
            if db.isRunning  then
                frame:SetEndlessTimer(self.db.timeElapsed, self.inviteTemplate)
                frame.contentContainer.timer:StartTimer()
                frame:Show()
            end
        elseif db.isRunning then
            frame:SetEndlessTimer(nil)
            frame.contentContainer.timer:SetTime(db.startTime, db.endTime)
            frame.contentContainer.timer:IsCounter(false)
            frame.contentContainer.timer:StartTimer()
            frame:Show()
        elseif db.durationTime then
            frame:SetEndlessTimer(nil)
            frame.contentContainer.timer:SetDuration(db.durationTime)
        end

        -- set session
        frame:SetSession(self)

        -- set zone filter and refresh zone
        frame.ZoneFilter = self.ZoneFilter
        frame:RefreshMap()

        -- set data
        self.mainFrame = frame
        frame.runningSession = self

        -- callback
        self:RegisterCallbackEvent("SessionTemplate_AddItem", Session.MainFrameOnNewItem)
        self:RegisterCallbackEvent("SessionTemplate_AddMoney", Session.MainFrameOnAddMoney)

        -- add into minimap button
        GGL.GUI.MainWindow:AddIntoMiniMapButton()
    else
        self:RemoveFromMainFrame()
        -- only hide mainFrame when session is running
        -- that way the frame is still open when the session ends but gets hidden if the state changes mid session
        if self.db.isRunning then
            GGL.GUI.MainWindow:CreateFrame():Hide()
        end
    end
end

function Session:RemoveFromMainFrame()
    if not self.mainFrame then return end
    local frame = self.mainFrame
    self.mainFrame = nil

    frame:SetSession(nil)

    frame.ZoneFilter = nil
    if self.db.isClosed then
        frame.contentContainer.timer:FinishTimer()
    else
        frame.contentContainer.timer:FinishTimer(true)
    end

    frame:RefreshMap()
    frame.runningSession = nil

    self:UnregisterCallbackEvent("SessionTemplate_AddItem", Session.MainFrameOnNewItem)
    self:UnregisterCallbackEvent("SessionTemplate_AddMoney", Session.MainFrameOnAddMoney)

    self.ownLogRebuilded = nil

    -- remove from minimap button
    GGL.GUI.MainWindow:RemoveFromMiniMapButton()
end

function Session:SetTimerData(startTime, endTime)
    if self.db.startTime then return end
    self.db.startTime = startTime
    self.db.endTime = endTime
    if endTime then
        self.db.durationTime = endTime - startTime
    end
end

function Session:SetTimerDuration(duration)
    local curTime = GGL.GetTime()
    self:SetTimerData(curTime, curTime + duration)
end

function Session:SetTimerEndless()
    self.db.endlessTime = true
    self.db.timeElapsed = self.db.timeElapsed or 0
    self.db.startTime = GGL.GetTime() - self.db.timeElapsed
end

local INI_ENTER_DATA_FORMAT, INI_RESET_DATA_FORMAT = "e:%d:%d", "r:%d:%d:%d"
local INI_SORT_SEARCH = ":%d+:(%d+)[:]*%d*"
local INI_DATA_SPLIT = "[er]+:(%d+):(%d+)[:]*(%d*)"
local function InstanceResetDataSorter(a, b)
    return a:match(INI_SORT_SEARCH) < b:match(INI_SORT_SEARCH)
end
function Session:UpdateResetData(playeTemplate, data, rawSet)
    if not playeTemplate then return end

    local playerMergeID = self:GetPlayerMergeID(playeTemplate)
    if not playerMergeID then return GGL:AddIntoErrorLog("Session:UpdateResetData / mergeID", playeTemplate.syncID) end

    -- create db data
    if data then
        -- merge data with db
        if not rawSet and self.db.instanceResetData[playerMergeID] then
            local newData = {}
            local db = self.db.instanceResetData[playerMergeID]
            local cache = {}
            for i = 1, #db do
                cache[#cache+1] = { db[i], str_split(":", db[i]) }
            end
            for i = 1, #data do
                local iType, zoneID, enterTime, resetTime = str_match(data[i], INI_DATA_SPLIT)
                local set = false
                for j = 1, #cache do
                    -- exact match
                    if cache[j][1] == data[i] then
                        break
                    -- zoneId + enterTime match / should be a new reset so replace
                    elseif cache[j][2] == zoneID and cache[j][3] == enterTime and resetTime then
                        newData[#newData+1] = data[i]
                        set = true
                        break
                    end
                end
                -- new record
                if not set then
                    newData[#newData+1] = data[i]
                end
            end
            -- resort table
            table.sort(newData, InstanceResetDataSorter)
            -- replace in db
            self.db.instanceResetData[playerMergeID] = newData
        else
            self.db.instanceResetData[playerMergeID] = data
        end
        self.db.instanceResetData[playerMergeID].time = GGL.GetTime()
        return self.db.instanceResetData[playerMergeID]
    end

    -- format own data
    local enterData, resetData
    if playeTemplate == self.selfPlayer then
        enterData = GGL:InstanceResets_GetInstanceEnterTimes(self:GetStartTime() - 1800, true)
        resetData = GGL:InstanceResets_GetInstanceResetTimes(self:GetStartTime())
    end

    local iniData = {}

    for i = 1, #enterData do
        iniData[#iniData + 1] = format(INI_ENTER_DATA_FORMAT, enterData[i].mapID or 0, enterData[i].time or 0)
    end

    for i = 1, #resetData do
        iniData[#iniData + 1] = format(INI_RESET_DATA_FORMAT, resetData[i].mapID or 0, resetData[i].enterTime or 0, resetData[i].startTime or 0)
    end

    -- resort table
    table.sort(iniData, InstanceResetDataSorter)

    -- set reset data for player
    return self:UpdateResetData(playeTemplate, iniData)
end

function Session:UpdateOwnResetData(noCallback)
    local data = self:UpdateResetData(self.selfPlayer)
    if not noCallback then
        self:FireCallbackEvent("SessionTemplate_OnInstancesResetUpdate", self, data)
    end
    return data
end

function Session:SetSpecialEvent(specialEvent)
    self.specialEvent = specialEvent
end

function Session:StartSpecialEvent(specialEvent)
    if self.specialEvent  == "Opening" then
        self.statistics:TrackOpeningItems(true)
        GGL.GUI.SpecialEvent:AddOpeningIntoMiniMapButton()
        GGL.GUI.SpecialEvent:ShowOpening()
        GGL.GUI.SpecialEvent:GetOpeningFrame():SetSession(self)
    end
end

function Session:SessionWasStarted()
    return self.db.wasStarted
end

function Session:Start()
    local db = self.db
    if db.isClosed or not db.dataIsSet then return end

    if not db.startTime then
        error(format("time not set for '%s'", self.sessionID))
    end

    -- statistics
    self.statistics:ResetSession()
    self:StartSpecialEvent()
    self.statistics:StartTracking()

    -- register events
    GGL.Loot:RegisterCallbackEvent("Loot_OnItemLoot", self.OnItemLoot_CB)
    GGL.Loot:RegisterCallbackEvent("Loot_OnMoneyLoot", self.OnMoneyLoot_CB)
    GGL:RegisterCallbackEvent("GGL_OnMapChange", self.OnMapChange_CB)
    GGL:RegisterCallbackEvent("GGL_OnPlayerKillNpc", self.OnPlayerKillNpc_CB)

    db.wasStarted = true
    db.isRunning = true
    db.isStopped = nil
    self.isRunning = true

    -- create the end timer
    if not self.timerUpdate and not self.db.endlessTime then
        self.timerUpdate = function()
            if db.endTime <= GGL.GetTime() then
                self:Close()
            else
                C_TimerAfter(TIMER_INTERVAL, self.timerUpdate)
            end
        end
        C_TimerAfter(TIMER_INTERVAL, self.timerUpdate)
    elseif self.db.endlessTime then
        self:SetTimerEndless()
        self.timerUpdate = function()
            if db.isClosed or not db.isRunning then
                return
            else
                db.timeElapsed = GGL.GetTime() - db.startTime
                C_TimerAfter(TIMER_INTERVAL, self.timerUpdate)
            end
        end
        C_TimerAfter(TIMER_INTERVAL, self.timerUpdate)
    end

    if self.mainFrame then
        self:AddIntoMainFrame(true)
        GGL.GUI.MainWindow:Show()
    end

    -- Instance resets
    if not self.cbOnInstancesResetUpdate then
        self.cbOnInstancesResetUpdate = function()
            self:UpdateOwnResetData()
        end
        GGL:RegisterCallbackEvent("GGL_OnInstancesResetNewLockout", self.cbOnInstancesResetUpdate)
        GGL:RegisterCallbackEvent("GGL_OnInstancesResetEnterInstance", self.cbOnInstancesResetUpdate)
    end

    -- rebuild list from db
    self:ReloadOwnItemsFromLog()

    -- local
    GGL.Session:LockSession(self)

    -- update zone for start
    self.OnMapChange_CB(nil, GGL:GetZoneData())

    -- init kill count
    if self.mainFrame then
        self.mainFrame:SetKillCount(db.ownKillCount or 0)
    end

    return db.startTime, db.endTime, db.durationTime
end

local function UnregisterAll(self)
    -- remove from loot callback
    GGL.Loot:UnregisterCallbackEvent("Loot_OnItemLoot", self.OnItemLoot_CB)
    GGL.Loot:UnregisterCallbackEvent("Loot_OnMoneyLoot", self.OnMoneyLoot_CB)
    GGL:UnregisterCallbackEvent("GGL_OnMapChange", self.OnMapChange_CB)
    GGL:UnregisterCallbackEvent("GGL_OnPlayerKillNpc", self.OnPlayerKillNpc_CB)

    -- stop statistics
    if self.statistics then
        self.statistics:StopTracking()
    end

    -- stop timer
    self:RemoveFromMainFrame()

    -- unregister all events
    self:UnregisterAllCallbackEvents()

    -- remove all special events
    GGL.GUI.SpecialEvent:RemoveAllFromFromMiniMapButton()

    -- reset events
    if self.cbOnInstancesResetUpdate then
        GGL:UnregisterCallbackEvent("GGL_OnInstancesResetNewLockout", self.cbOnInstancesResetUpdate)
        GGL:UnregisterCallbackEvent("GGL_OnInstancesResetEnterInstance", self.cbOnInstancesResetUpdate)
        self.cbOnInstancesResetUpdate = nil
    end

    -- remove from running sessions
    RunningSessions[self.db.sessionID] = nil
end

function Session:Unregister()
    UnregisterAll(self)
end

function Session:Stop()
    -- only stop runnig sessions
    if not self.db.isRunning then return end
    local db = self.db
    db.isStopped = true
    db.isRunning = false
    self.isRunning = false

    self:FireCallbackEvent("SessionTemplate_SessionStopped", self)

    UnregisterAll(self)

    -- unlock
    GGL.Session:UnlockSession(self)

    RunningSessions[self.db.sessionID] = self
end

--- Close the session
function Session:Close()
    local db = self.db
    db.isClosed = true
    db.isRunning = false
    db.isStopped = nil
    self.isRunning = false

    db.closeTime = GGL.GetTime()

    self:FireCallbackEvent("SessionTemplate_SessionClosed", self)

    -- move and delete link
    if self.dbRaw[db.sessionID] then
        GGL:AddSessionIntoArchive(self.dbRaw[db.sessionID])
        self.dbRaw[db.sessionID] = nil
    end

    UnregisterAll(self)

    -- unlock
    GGL.Session:UnlockSession(self)
end

-- ##################
-- Rebuild from itemLog
-- ##################
local SESSION_REBUILD_ID = "self"

local function ProcessLogItemEventCB(session, state, currentRun, numRuns, timeLeftInSec, rebuildID)
    if rebuildID ~= SESSION_REBUILD_ID then return end
    if state == "start" then
        GGL:FormatPrint("Start rebuild of '%s' with %d steps. Rebuild time %.1fsec.", rebuildID, numRuns, timeLeftInSec)
    elseif state == "run" then
        --GGL:FormatPrint("Run rebuild of '%s' with step %d/%d. Time left %.1fsec.", rebuildID, currentRun, numRuns, timeLeftInSec)
    elseif state == "end" then
        GGL:FormatPrint("Rebuild of '%s' done.", rebuildID)
        session:UnregisterCallbackEvent("LogRebuild_StatusChange", ProcessLogItemEventCB)
    end
end

local function ProcessLogItem(session, logEntry)
    session:AddFromLogString(logEntry[2])
end

function Session:ReloadOwnItemsFromLog()
    if self.ownLogRebuilded then return end

    -- first get a full list of all item log entrys
    local rebuildList = {}
    for playerSyncID, lootData in pairs(self.db.lootLog) do
        -- ignore empty tables
        if #lootData > 0 then
            local player = Templates.Sync.GetPlayerFromSync(playerSyncID, true)
            -- only player
            if player.isPlayerBattleTag then
                for i = 1, #lootData do
                    --self:AddFromLogString(lootData[i])
                    rebuildList[#rebuildList+1] = { player, lootData[i] }
                end
            end
        end
    end

    self:RegisterCallbackEvent("LogRebuild_StatusChange", ProcessLogItemEventCB)

    GGL:LogRebuild_ProcessLogList(self, rebuildList, ProcessLogItem, SESSION_REBUILD_ID)

    self.ownLogRebuilded = true
end

-- ##################
-- Helper
-- ##################
function Session:GetType()
    return SessionHelper:GetType()
end

function Session:IsRunning()
    return self.db.isRunning
end

function Session:IsStopped()
    return self.db.isStopped and true or false
end

function Session:IsRunningTemplate()
    return self.isRunning
end

function Session:IsClosed()
    return self.db.isClosed
end

function Session:GetSet()
    return self.db.set
end

function Session:GetSessionType()
    return self.sessionType
end

function Session:DataIsSet()
    return self.db.dataIsSet
end

function Session:GetTimer()
    return self.db.startTime, self.db.endTime, self.db.durationTime
end

function Session:GetStartTime()
    return self.db.startTime
end

--- Returns session end time
-- For endless sessions that returns 0
function Session:GetEndTime()
    return self.db.endlessTime and 0 or self.db.endTime
end

--- Returns session duration
-- For endless sessions that returns the startTime
function Session:GetDuration()
    return self.db.endlessTime and self.db.startTime or self.db.durationTime
end

function Session:GetSessionID()
    return self.db.sessionID
end

function Session:LoggedItemUseVariations(itemID, itemString)
    if self.ItemPrice then
        return self.ItemPrice:LoggedItemUseVariations(itemID, itemString)
    end
end